@@ -13,25 +13,26 @@ written permission is obtained from ElasticBox
 '''
 
 import argparse
-import getpass
 import httplib
 import inspect
 import json
+import keyring
 import os
 import requests
-import stat
 import sys
 import traceback
 import urlparse
 
 from mimetypes import types_map
 
+
 REQUESTS_TIMEOUT_SECONDS = 30
 
-CONFIG_PATH = os.path.expanduser('~/')
-DEFAULT_TOKEN_PATH = os.path.join(CONFIG_PATH, '.eb_token')
+EBX_CREDENTIALS_NAME = 'ElasticBox Token'
+EBX_CREDENTIALS_ACCOUNT = 'elasticbox'
 
-DEPLOY_SERVICE_REQUEST_SCHEMA = 'http://elasticbox.net/schemas/2014-06-04/deploy-instance-request'
+SCHEMA_VERSION = '2014-10-09'
+DEPLOY_SERVICE_REQUEST_SCHEMA = 'http://elasticbox.net/schemas/{0}/deploy-instance-request'.format(SCHEMA_VERSION)
 
 TOKEN_URL = '/services/security/token'
 
@@ -83,10 +84,6 @@ TAGS_PARAM = (
      'nargs': '*',
      'required': False}
 )
-PASSWORD_PARAM = (
-    ['--password'],
-    {'help': 'Password', 'required': False}
-)
 SCRIPT_NAME_PARAM = (
     ['-s', '--script-name'],
     {'help': 'Event script name', 'required': True}
@@ -118,9 +115,8 @@ ADDRESS_PARAM = (
      'default': DEFAULT_URL}
 )
 TOKEN_PARAM = (
-    ['-token', '--token_file'],
-    {'help': '',
-     'default': DEFAULT_TOKEN_PATH}
+    ['--token'],
+    {'help': 'Authentication Token'}
 )
 BOX_ID_PARAM = (
     ['-b', '--box-id'],
@@ -132,6 +128,12 @@ WORKSPACE_ID_PARAM = (
     {'help': 'Workspace Id',
      'required': True}
 )
+NO_KEYCHAIN_PARAM = (
+    ['--no-keychain'],
+    {'action': 'store_true',
+     'required': False,
+     'help': 'Disable the use of keychain service to store your Authentication Token'}
+)
 
 
 def _generate_resource_id_param(required, help_text):
@@ -167,17 +169,6 @@ def _generate_field_param(help_text):
     )
 
 
-def _generate_email_param(required):
-    if required:
-        commands = ['-e', '--email']
-    else:
-        commands = ['--email']
-    return (
-        commands,
-        {'help': 'Email', 'required': required}
-    )
-
-
 class UnexpectedFieldsException(Exception):
     pass
 
@@ -190,7 +181,7 @@ class ArgumentException(Exception):
     pass
 
 
-class InvalidTokenException(Exception):
+class AuthTokenException(Exception):
     pass
 
 
@@ -199,7 +190,7 @@ class AmbigousResourceFilterException(Exception):
 
 
 class _BaseCommand(object):
-    """Base class for all eb.py commands."""
+    """Base class for all commands."""
 
     def __init__(self, args):
         self.args = args
@@ -229,17 +220,16 @@ class _BaseCommand(object):
                 print resource
 
     def check_login(self):
-        if not self.args.email and not self.args.password:
-            try:
-                client.read_token_from_file(DEFAULT_TOKEN_PATH)
-                return
-            except IOError:
-                raise InvalidTokenException('You need to log in before calling any functions')
-        elif self.args.email and not self.args.password:
-            self.args.password = getpass.getpass('Password: ')
+        if not self.args.token and not self.args.no_keychain:
+            client.login_with_token()
+            return
+        if not self.args.token:
+            self.args.token = raw_input('Authentication Token: ')
+            if not self.args.token:
+                raise AuthTokenException("You need to get your token before calling any functions")
 
         address = DEFAULT_URL if self.args.address is None else self.args.address
-        client.log_in(address, self.args.email, self.args.password)
+        client.log_in(address, self.args.token, self.args.no_keychain)
 
     def _extract_fields(self, resource, fields):
         if len(set(fields) - set(resource.keys())) != 0:
@@ -374,20 +364,39 @@ class _RESTBase(object):
 
 class LoginCommand(_BaseCommand):
     def login(self):
-        if not self.args.password:
-            self.args.password = getpass.getpass('Password: ')
+        token = self.args.token
+        address = self.args.address
+        if (not self.args.token and
+                not self.args.no_keychain and
+                keyring.get_password(EBX_CREDENTIALS_NAME, EBX_CREDENTIALS_ACCOUNT) is not None):
 
-        client.log_in(self.args.address, self.args.email, self.args.password)
-        if not os.path.exists(CONFIG_PATH):
-            os.makedirs(CONFIG_PATH)
+            credentials = keyring.get_password(EBX_CREDENTIALS_NAME, EBX_CREDENTIALS_ACCOUNT)
+            credentials = credentials.split(",")
+            token = credentials[0]
+            address = credentials[1]
 
-        client.write_token_to_file(token_file_path=self.args.token_file)
-        print 'Login successful, token saved on {}'.format(self.args.token_file)
+        elif ((not self.args.token and self.args.no_keychain) or
+              (not self.args.token and not self.args.no_keychain and
+               keyring.get_password(EBX_CREDENTIALS_NAME, EBX_CREDENTIALS_ACCOUNT) is None)):
+            token = raw_input('Authentication Token: ')
+            if not token:
+                # Lets re-define the way the user your surf to our documenation or elasticbox.com website.
+                print 'You need a token to access your elasticbox account. Log in to your elasticbox website ' \
+                    'and create a token from Authentication Tokens under your profile drop-down menu.'
+                raise AuthTokenException("You need to get your token before calling any functions")
+
+        client.log_in(address, token, self.args.no_keychain)
+
+        if not self.args.no_keychain:
+            print 'Login successful on {0}, token saved in your keychain service as {1}'.format(address,
+                                                                                                EBX_CREDENTIALS_NAME)
+        else:
+            print 'Login successful using token {0} on {1}'.format(token, address)
 
     @staticmethod
     def parser(subparsers):
-        _subparser(subparsers, 'login', [_generate_email_param(True), PASSWORD_PARAM, ADDRESS_PARAM, TOKEN_PARAM],
-                   help_text='Log in to ElasticBox and save the token')
+        _subparser(subparsers, 'login', [ADDRESS_PARAM, TOKEN_PARAM, NO_KEYCHAIN_PARAM],
+                   help_text='Log in to ElasticBox and get your Authentication token')
 
 
 class BoxesCommand(_BaseCommand):
@@ -440,33 +449,32 @@ class BoxesCommand(_BaseCommand):
 
         _subparser(box_subparser, 'list',
                    [_generate_resource_id_param(False, 'Workspace Id'),
-                    _generate_resource_name_param(False, 'Box name'), TAGS_PARAM, _generate_email_param(False),
-                    PASSWORD_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM],
+                    _generate_resource_name_param(False, 'Box name'), TAGS_PARAM, TOKEN_PARAM, NO_KEYCHAIN_PARAM,
+                    ADDRESS_PARAM, JSON_OUTPUT_PARAM],
                    help_text='List of existing boxes')
         _subparser(box_subparser,
                    'get',
                    [_generate_resource_id_param(True, 'Box Id'), _generate_field_param(field_help_text),
-                    _generate_email_param(False), PASSWORD_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM],
+                    TOKEN_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Get specified fields of given box. Available fields are: id, name, service,'
                    ' organizations, tags, variables, created, uri, events, members, owner, bindings, schema')
         _subparser(box_subparser,
                    'create',
-                   [INPUT_FILE_PARAM, _generate_email_param(False), PASSWORD_PARAM, ADDRESS_PARAM],
+                   [INPUT_FILE_PARAM, TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Create box from json definition')
         _subparser(box_subparser,
                    'set',
-                   [_generate_resource_id_param(True, 'Box Id'), VARIABLE_PARAM, _generate_email_param(False),
-                    PASSWORD_PARAM, ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Box Id'), VARIABLE_PARAM, TOKEN_PARAM, NO_KEYCHAIN_PARAM,
+                    ADDRESS_PARAM],
                    help_text='Update a variable of the given box')
         _subparser(box_subparser,
                    'delete',
-                   [_generate_resource_id_param(True, 'Box Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Box Id'), TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Delete given box')
         _subparser(box_subparser,
                    'upload_script',
-                   [_generate_resource_id_param(True, 'Box Id'), SCRIPT_NAME_PARAM, INPUT_FILE_PARAM,
-                    _generate_email_param(False), PASSWORD_PARAM, ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Box Id'), SCRIPT_NAME_PARAM, INPUT_FILE_PARAM, NO_KEYCHAIN_PARAM,
+                    TOKEN_PARAM, ADDRESS_PARAM],
                    help_text='Upload script to given box')
 
 
@@ -505,23 +513,22 @@ class ProfilesCommand(_BaseCommand):
 
         _subparser(profile_subparser,
                    'create',
-                   [INPUT_FILE_PARAM, _generate_email_param(False), PASSWORD_PARAM, ADDRESS_PARAM],
+                   [INPUT_FILE_PARAM, TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Creates a new profile from a json file')
         _subparser(profile_subparser,
                    'get',
                    [_generate_resource_id_param(True, 'Profile Id'), _generate_field_param(help_field_text),
-                    _generate_email_param(False), PASSWORD_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM],
+                    TOKEN_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Gets specified fields of given deployment profile. Available fields are: id, name, box,'
                    ' updated, created, uri, instances, members, owner, schema')
         _subparser(profile_subparser,
                    'list',
                    [_generate_resource_id_param(False, 'Workspace Id'), _generate_resource_name_param(True, 'Box name'),
-                    _generate_email_param(False), PASSWORD_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM],
+                    TOKEN_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Lists all deployment profiles of a box')
         _subparser(profile_subparser,
                    'delete',
-                   [_generate_resource_id_param(True, 'Profile Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Profile Id'), TOKEN_PARAM, NO_KEYCHAIN_PARAM, ADDRESS_PARAM],
                    help_text='Deletes given profile')
 
 
@@ -593,64 +600,56 @@ class InstancesCommand(_BaseCommand):
                    'list',
                    [_generate_resource_id_param(False, 'Workspace Id'),
                     _generate_resource_name_param(False, 'Box name'), _generate_tag_param(False), SERVICE_ID_PARAM,
-                    _generate_email_param(False), PASSWORD_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM],
+                    TOKEN_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Lists of existing instances')
         _subparser(instance_subparser,
                    'check_status',
-                   [_generate_resource_id_param(True, 'Instance Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Instance Id'), TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Check status of given instance')
         _subparser(instance_subparser,
                    'get',
                    [_generate_resource_id_param(True, 'Instance Id'), _generate_field_param(field_help_text),
-                    _generate_email_param(False), PASSWORD_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM],
+                    TOKEN_PARAM, ADDRESS_PARAM, JSON_OUTPUT_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Gets specified fields of given instance. Available fields are: id, name, updated, '
                    'created, service, tags, variables, uri, boxes, operation, state, members, owner, bindings, schema')
         _subparser(instance_subparser,
                    'deploy',
                    [PROFILE_NAME_PARAM, WORKSPACE_ID_PARAM, BOX_ID_PARAM,
-                    _generate_resource_name_param(False, 'Box name'), TAGS_PARAM, _generate_email_param(False),
-                    PASSWORD_PARAM, ADDRESS_PARAM],
+                    _generate_resource_name_param(False, 'Box name'), TAGS_PARAM, TOKEN_PARAM, NO_KEYCHAIN_PARAM,
+                    ADDRESS_PARAM],
                    help_text='Deploys given box with given deployment profile')
         _subparser(instance_subparser,
                    'power_on',
-                   [_generate_resource_id_param(True, 'Instance Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Instance Id'), TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Power on given instance')
         _subparser(instance_subparser,
                    'shut_down',
-                   [_generate_resource_id_param(True, 'Instance Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Instance Id'), TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Shutdown given instance')
         _subparser(instance_subparser,
                    'reconfigure',
-                   [_generate_resource_id_param(True, 'Instance Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Instance Id'), TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Reconfigure given instance')
         _subparser(instance_subparser,
                    'reinstall',
-                   [_generate_resource_id_param(True, 'Instance Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Instance Id'), TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Reinstall given instance')
         _subparser(instance_subparser,
                    'terminate',
-                   [_generate_resource_id_param(True, 'Instance Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Instance Id'), TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Terminates given instance')
         _subparser(instance_subparser,
                    'force_terminate',
-                   [_generate_resource_id_param(True, 'Instance Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Instance Id'), TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Forces terminate given instance')
         _subparser(instance_subparser,
                    'delete',
-                   [_generate_resource_id_param(True, 'Instance Id'), _generate_email_param(False), PASSWORD_PARAM,
-                    ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Instance Id'), TOKEN_PARAM, ADDRESS_PARAM, NO_KEYCHAIN_PARAM],
                    help_text='Deletes given instance')
         _subparser(instance_subparser,
                    'set',
-                   [_generate_resource_id_param(True, 'Instance Id'), VARIABLE_PARAM, _generate_email_param(False),
-                    PASSWORD_PARAM, ADDRESS_PARAM],
+                   [_generate_resource_id_param(True, 'Instance Id'), VARIABLE_PARAM, TOKEN_PARAM, NO_KEYCHAIN_PARAM,
+                    ADDRESS_PARAM],
                    help_text='Updates a variable of the given instance')
 
 
@@ -947,7 +946,8 @@ class ElasticBoxClient(object):
 Example usage:
 
     client = ElasticBoxClient()
-    client.log_in('localhost', 'admin@elasticbox.com', 'elasticbox')
+    ## To login you may have your EB Authentication Token.
+    client.log_in('localhost', '7c94811c-dd3f-4a7f-a5db-edb2c9a358af')
 
     # Delete all instances
     for instance in client.instances.get_all():
@@ -970,61 +970,82 @@ Example usage:
         self.token = None
         self.host_address = None
         self.controller_routes = None
+        self.no_keychain = False
         # We actually read the routes to the web service controllers from the web services itself.
         # So we don't assume that the route to the instances controller is services/instances; instead
         # we query services/roots to tell us what all the other routes are.
-        # You can see this in either read_token_from_file in log_in below.
+        # You can see this in log_in below.
 
     def _generate_header(self, token):
         return {EB_TOKEN_HEADER: token, USER_AGENT_HEADER: USER_AGENT_EB}
 
-    def read_token_from_file(self, token_file_path=DEFAULT_TOKEN_PATH):
-        # The client needs a token to issue calls to the web services. You can either
-        # read this token from a file (this method) or log in directly with a password
-        # and username (the log_in method below).
-        if not os.path.exists(token_file_path):
-            raise IOError('Token file specified doesn\'t exist')
-
+    def login_with_token(self):
         try:
-            with open(token_file_path) as token_file:
-                info = token_file.read().split(',')
-                self.token = info[0]
-                self.host_address = info[1]
+            if not self.no_keychain:
+                self.token, self.host_address = self.read_token_from_keychain()
 
             headers = self._generate_header(self.token)
             controller_routes_response = requests.get(
                 '{}/services/roots'.format(self.host_address), headers=headers, verify=False)
             controller_routes_response.raise_for_status()
             if controller_routes_response.status_code in [httplib.FORBIDDEN, httplib.UNAUTHORIZED]:
-                raise InvalidTokenException('Authenticating with token from file {} failed.'
-                                            ' Try logging in manually with login command'.format(token_file_path))
+                raise AuthTokenException("Authenticating with token from keychain {} failed."
+                                         " Try logging in manually with login command")
 
             self.controller_routes = dict()
             for key, value in controller_routes_response.json().items():
                 self.controller_routes[str(key)] = '{}/'.format(value.strip('/'))
-        except:
-            raise InvalidTokenException("File {0} does not contain a valid token.".format(token_file_path))
+        except ResourceNotFound:
+            raise
+        except Exception:
+            if not self.no_keychain:
+                raise AuthTokenException("Keychain service does not contain a valid "
+                                         "{0}.".format(EBX_CREDENTIALS_NAME))
+            else:
+                raise AuthTokenException("This token {} is not valid".format(self.token))
 
-    def write_token_to_file(self, token_file_path=DEFAULT_TOKEN_PATH):
-        with open(token_file_path, 'w') as token_file:
-            token_file.write('{},{}'.format(self.token, self.host_address))
-        os.chmod(token_file_path, stat.S_IREAD | stat.S_IWRITE)
+    def read_token_from_keychain(self):
+        credentials = keyring.get_password(EBX_CREDENTIALS_NAME, EBX_CREDENTIALS_ACCOUNT)
+        if not credentials:
+            raise AuthTokenException("ElasticBox Token doesn't exist in your keychain service")
 
-    def log_in(self, host_address, email, password):
-        payload = dict(email=email, password=password, remember='on')
+        credentials = credentials.split(",")
+        return (credentials[0], credentials[1])
+
+    def log_in(self, host_address, token, no_keychain):
+        payload = dict(token=token, remember='on')
         if not (host_address.startswith('http://') or host_address.startswith('https://')):
             revised_address = 'http://{}'.format(host_address.strip('/'))
         else:
             revised_address = host_address
 
-        response = requests.post(
-            urlparse.urljoin(revised_address, TOKEN_URL),
-            data=payload,
-            verify=False)
-        response.raise_for_status()
-        token = str(response.text).strip('"')
+        try:
+            response = requests.post(
+                urlparse.urljoin(revised_address, TOKEN_URL),
+                data=payload,
+                verify=False)
+            response.raise_for_status()
+        except requests.exceptions.HTTPError as error:
+            if response.status_code in [httplib.FORBIDDEN, httplib.UNAUTHORIZED]:
+                raise AuthTokenException("Failed to authenticate with token {0} because it's invalid or "
+                                         "expired.".format(token))
+            else:
+                raise AuthTokenException(error.message)
+        except requests.exceptions.ConnectionError as error:
+            # nodename nor servname provided, or not known
+            if error.args and error.args[0].reason and error.args[0].reason.errno == 8:
+                raise AuthTokenException("Can't connect to {0}. Verify you can connect to the network and provide "
+                                         "the right address.".format(revised_address))
+            else:
+                raise AuthTokenException(error.message)
+
         self.token = token
         self.host_address = revised_address
+        self.no_keychain = no_keychain
+
+        if not self.no_keychain:
+            keyring.set_password(EBX_CREDENTIALS_NAME, EBX_CREDENTIALS_ACCOUNT, '{0},{1}'.format(self.token,
+                                                                                                 self.host_address))
 
         headers = self._generate_header(self.token)
         raw_controller_info = requests.get(
@@ -1103,21 +1124,22 @@ def _generate_exit_code(exit_code, message):
 def start():
     try:
         _parse_command_line(sys.argv[1:])
+    except AuthTokenException as error:
+        print error.message
     except requests.exceptions.HTTPError as error:
         code = 1
         if error.response is not None:
             code = EXIT_CODES[error.response.status_code]
         _generate_exit_code(code, 'URL: {0} Response: {1}'.format(error.response.url, error.message))
-    except InvalidTokenException as error:
-        _generate_exit_code(10, error.message)
-    except UnexpectedFieldsException as error:
-        _generate_exit_code(11, error.message)
     except ResourceNotFound as error:
         _generate_exit_code(6, error.message)
+    except UnexpectedFieldsException as error:
+        _generate_exit_code(11, error.message)
     except IOError as error:
         _generate_exit_code(12, error.message)
     except Exception as error:
         _generate_exit_code(1, error.message)
 
+
 if __name__ == '__main__':
     sys.exit(start())
